import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide" # hide pygame welcome message
import pygame
from pygame import mixer
import random
import math
import time
from io import StringIO
from PIL import Image

print("""
########   ######
##     ## ##    ##    ##     ##
##     ## ##          ##     ##
########  ##   #### ###### ######
##        ##    ##    ##     ##
##        ##    ##    ##     ##
##         ######



Built arround Pygame -- (Pg++) -- V0.2.3
""")
pygame.init()
pygame.font.init()

# colors
BLACK = (0, 0, 0)

BG = BLACK

# elements
elements = {}
delays = {}
animations = {}
particles = {}

# var
isButtonKPressed = False
devMode = False
test = None

# create window
def createWindow(title="Py++", width=600, height=400, icon=None):
    global WWIDTH
    global WHEIGHT
    WWIDTH = width
    WHEIGHT = height
    global WIN
    WIN = pygame.display.set_mode((width, height))
    pygame.display.set_caption(title)
    pygame.display.set_icon(pygame.image.load(icon))
    return WIN

# start a clock object to cap fps
def startClock():
    global clock
    clock =  pygame.time.Clock()

# fps clock
def clockTick(fps=30):
    global clock
    clock.tick(fps)

# set background variable
def setBackground(color=BLACK):
    global BG
    BG = color

## dev mode
def turnOnDevMode():
    global devMode
    devMode = True

def turnOffDevMode():
    global devMode
    devMode = False

def toggleDevMode():
    global devMode
    devMode = not devMode

# return pygame events
def getEvents():
    return pygame.event.get()

# pygame quit event
def eventQuit():
    return pygame.QUIT

# single key press detection
def newKeyPressed():
    return pygame.KEYDOWN

# checks what keys are pressed that frame
def getKeyPressed():
    return pygame.key.get_pressed()

# get mouse position
def getMousePos():
    x,y = pygame.mouse.get_pos()
    return x, y

# get mouse buttons clicked by indexing it
def mouseClick():
    return pygame.mouse.get_pressed()

# background music
def startAudio(path, volume=100, loop=-1):
    mixer.music.load(path)
    mixer.music.set_volume(volume/100)
    mixer.music.play(loop)

# sound effects
def playAudio(path):
    sound = mixer.Sound(path)
    sound.play()

### keys
def keyA():
    return pygame.K_a
def keyB():
    return pygame.K_b
def keyC():
    return pygame.K_c
def keyD():
    return pygame.K_d
def keyE():
    return pygame.K_e
def keyF():
    return pygame.K_f
def keyG():
    return pygame.K_g
def keyH():
    return pygame.K_h
def keyI():
    return pygame.K_i
def keyJ():
    return pygame.K_j
def keyK():
    return pygame.K_k
def keyL():
    return pygame.K_l
def keyM():
    return pygame.K_m
def keyN():
    return pygame.K_n
def keyO():
    return pygame.K_o
def keyP():
    return pygame.K_p
def keyQ():
    return pygame.K_q
def keyR():
    return pygame.K_r
def keyS():
    return pygame.K_s
def keyT():
    return pygame.K_t
def keyU():
    return pygame.K_u
def keyV():
    return pygame.K_v
def keyW():
    return pygame.K_w
def keyX():
    return pygame.K_x
def keyY():
    return pygame.K_y
def keyZ():
    return pygame.K_z
def keyUP():
    return pygame.K_UP
def keyLEFT():
    return pygame.K_LEFT
def keyDOWN():
    return pygame.K_DOWN
def keyRIGHT():
    return pygame.K_RIGHT
def keyTAB():
    return pygame.K_TAB
def keySHIFT():
    return pygame.K_LSHIFT
def keyESC():
    return pygame.K_ESCAPE
def keySPACE():
    return pygame.K_SPACE
def keyCTRL():
    return pygame.K_LCTRL
def key0():
    return pygame.K_0
def key1():
    return pygame.K_1
def key2():
    return pygame.K_2
def key3():
    return pygame.K_3
def key4():
    return pygame.K_4
def key5():
    return pygame.K_5
def key6():
    return pygame.K_6
def key7():
    return pygame.K_7
def key8():
    return pygame.K_8
def key9():
    return pygame.K_9

# quit the game
def gameQuit():
    pygame.quit()

# window refresh
def drawWindow():
    global WIN
    global BG
    global devMode
    global test
    WIN.fill(BG)

    # mouse detection
    mx, my = getMousePos()
    elements["mouse"] = {"id": "mouse", "type": "mouse", "rect": pygame.Rect(mx, my, 1, 1)}

    for sprite in elements.values(): # draw elements
        if sprite["type"] == "sprite":
            item = pygame.transform.flip(pygame.transform.rotate(pygame.transform.scale(sprite["image"], (sprite["sx"], sprite["sy"])), sprite["r"]), sprite["f"][0], sprite["f"][1])
            x = sprite["x"] - item.get_width()/2
            y = sprite["y"] - item.get_height()/2
            sprite["rect"].x = sprite["x"] - sprite["rect"].w/2
            sprite["rect"].y = sprite["y"] - sprite["rect"].h/2
            WIN.blit(item, (x, y))
            if devMode:
                font = pygame.font.Font(None, 20)
                text = font.render(f"x:{x}|y:{y}|sx:{item.get_width()}|sy:{item.get_height()}", True, (0, 0, 0))
                pygame.draw.rect(WIN, (0, 0, 0), sprite["rect"], 2)
                WIN.blit(text, (sprite["rect"].x, sprite["rect"].y-25))
        elif sprite["type"] == "text":
            if devMode:
                font = pygame.font.Font(None, 20)
                text = font.render("x:"+str(sprite["rect"].x)+"|y:"+str(sprite["rect"].y)+"|sx:"+str(sprite["rect"].w)+"|sy:"+str(sprite["rect"].h), True, (0, 0, 0))
                pygame.draw.rect(WIN, (0, 0, 0), sprite["rect"], 2)
                WIN.blit(text, (sprite["rect"].x, sprite["rect"].y-25))
            WIN.blit(sprite["text"], sprite["rect"])
    pygame.display.update()

# create a new element
def setElement(id, image, x=0, y=0, scale=100, col_x=100, col_y=100, rotation=0, flip=[False, False], colision_type=[1], group=1):
    if image != "none":
        image = pygame.image.load(image).convert_alpha()
        #image.set_colorkey(None)
        w, h = image.get_width(), image.get_height()
        size_x, size_y = scale*w/100, scale*h/100
    else:
        image = None
    elements[id] = {"id":id, "type": "sprite", "group":group, "image":image, "rect":pygame.Rect(x, y, col_x, col_y), "scale": scale, "x":x, "y":y, "sx":size_x, "sy":size_y, "r":rotation, "f":flip, "c":colision_type}

# create a text element
def writeText(id, string, font, coordx, coordy, fontSize, color=(200, 200, 200), group=1): #Function to set text
    font = pygame.font.Font(font, fontSize)
    text = font.render(string, True, color)
    textRect = text.get_rect()
    textRect.center = (coordx, coordy)
    elements[id] = {"id": id, "type": "text", "group":group, "text": text, "rect": textRect}

# check if mouse click on a element
def isElementPressed(button):
    global isButtonKPressed
    if isButtonKPressed == False:
        if isColiding("mouse", button) and mouseClick()[0]:
            isButtonKPressed = True
            return True
        else:
            return False
    elif isButtonKPressed == True and not mouseClick()[0]:
        isButtonKPressed = False
        return False
    else:
        return False

# set element size
def setElementSizeX(id, x):
    elements["id"]["sx"] = x

def setElementSizeY(id, y):
    elements["id"]["sy"] = y

# move sprite
def moveXSprite(id, x):
    elements[id]["x"] += x

def moveYSprite(id, y):
    elements[id]["y"] += y

# try to move sprite with colisions
def tryMoveX(id, x, bxm=None, bxM=None):
    # check if touching window border
    if elements[id]["x"] - elements[id]["sx"]/2 + x >= 0 and elements[id]["x"] + elements[id]["sx"]/2 + x <= WWIDTH:
        if elements[id]["c"] != [0]: # check if the player is colision
            elements[id]["rect"].x += x
            for sprite in elements.values():
                if not sprite == elements[id]: # if target its not itself
                    if not sprite["c"] == [0]: # if target has colision
                        for group in elements[id]["c"]: # get colision group
                            if group in sprite["c"]: # if the colision group matches
                                if pygame.Rect.colliderect(sprite["rect"], elements[id]["rect"]):
                                    elements[id]["rect"].x += x*-1
                                    return
        moveXSprite(id, x)

def tryMoveY(id, y, bym=None, byM=None):
    if elements[id]["y"] - elements[id]["sy"]/2 + y >= 0 and elements[id]["y"] + elements[id]["sy"]/2 + y <= WHEIGHT:
        if elements[id]["c"] != [0]: # check if the player is colision
            elements[id]["rect"].y += y
            for sprite in elements.values():
                if not sprite == elements[id]: # if target its not itself
                    if not sprite["c"] == [0]: # if target has colision
                        for group in elements[id]["c"]: # get colision group
                            if group in sprite["c"]: # if the colision group matches
                                if pygame.Rect.colliderect(sprite["rect"], elements[id]["rect"]):
                                    elements[id]["rect"].y += y*-1
                                    return
        moveYSprite(id, y)

# set sprite position
def setXSprite(id, x):
    elements[id]["x"] = x

def setYSprite(id, y):
    elements[id]["y"] = y

# move in a direction
def moveInDirection(id, r, steps):
    x = math.sin(math.radians(r)) * steps
    y = math.cos(math.radians(r)) * steps
    moveXSprite(id, x*-1)
    moveYSprite(id, y*-1)

# try to move towards a point
def tryMoveInDirection(id, r, steps):
    x = math.sin(math.radians(r)) * steps
    y = math.cos(math.radians(r)) * steps
    tryMoveX(id, x*-1)
    tryMoveY(id, y*-1)

# move towards a target
def moveTowards(id, target, steps):
    angle = math.degrees(math.atan2(elements[target]["x"] - elements[id]["x"], elements[target]["y"] - elements[id]["y"])) + 180
    moveInDirection(id, angle, steps)

# get sprite info
def getXSprite(id):
    return elements[id]["x"]

def getYSprite(id):
    return elements[id]["y"]

def getRotation(id):
    return elements[id]["r"]

# point a element towards a direction
def point(id, angle):
    elements[id]["r"] = angle

# rotate some degrees in a direction
def turn(id, degrees):
    elements[id]["r"] += degrees

# point towards a target
def pointTowards(id, target):
    point(id, math.degrees(math.atan2(elements[target]["x"] - elements[id]["x"], elements[target]["y"] - elements[id]["y"])) + 180)

# point towards mouse pointer
def pointTowardsMouse(id):
    x, y = getMousePos()
    point(id, math.degrees(math.atan2(x - elements[id]["x"], y - elements[id]["y"])) + 180)

# checks if the element is past the window border
def isTouchingBorder(id):
    if elements[id]["x"] - elements[id]["sx"]/2 >= 0 and elements[id]["x"] + elements[id]["sx"]/2 <= WWIDTH:
        if elements[id]["y"] - elements[id]["sy"]/2 >= 0 and elements[id]["y"] + elements[id]["sy"]/2 <= WHEIGHT:
            return False
    return True

# checks if two elements are coliding
def isColiding(col1, col2):
    return pygame.Rect.colliderect(elements[col1]["rect"], elements[col2]["rect"])

# checks if its coliding with any other element in its group
def isColidingGroup(id, tgroup):
    for sprite in elements.values(): # get all elements
        if not sprite == elements[id]: # if target its not itself
            if not sprite["c"] == [0]: # if target has colision
                for group in tgroup: # get colision group
                    if group in sprite["c"]: # if the colision group matches
                        if pygame.Rect.colliderect(sprite["rect"], elements[id]["rect"]):
                            return True
    return False

# delete element
def delElement(id):
    try:
        del elements[id]
    except:
        pass

# create a delay object that will return true when its time has passed
def delay(id, seconds):
    if id in delays:
        if time.time() - delays[id][2] >= delays[id][1]:
            del delays[id]
            return True
        else:
            return False
    else:
        delays[id] = [id, float(seconds), time.time()]

# set element ghost effect
def setGhostEffect(id, effect):
    elements[id]["image"].set_alpha(effect)

# print all elements in layer order
def showLayers():
    for idx, v in enumerate(elements):
        print(str(idx+1) + " -" + "-"*idx + " " + v)

# random
def pickRandom(start, end):
    return random.randint(start, end)

# set image
def setImage(id, image):
    elements[id]["image"] = pygame.image.load(image).convert_alpha()

# animate
def animate(id, anims, delay_):
    if id in animations:
        if delay(id, delay_):
            if animations[id]["currentAnim"] >= len(anims)-1:
                animations[id]["currentAnim"] = 0
            else:
                animations[id]["currentAnim"] += 1
            setImage(id, animations[id]["anims"][animations[id]["currentAnim"]])
    else:
        animations[id] = {"id": id, "anims": anims, "currentAnim": 0, "delay": delay_}

# flip
def flip(id, flip=[False, False]):
    elements[id]["f"] = flip

# add particles
def particles(id, textures, radius, density, type_):
    if type_ == "rain":
        pass
